/*************************************************************************************/
/*--Copyright (c) 2015 4.7 by SAS Institute Inc., Cary, NC, USA. All Rights Reserved.*/ 
/*                                                                                   */ 
/*Model         BANKING DDS 4.7                                                      */ 
/*Model Version 4.7                                                                  */ 
/*Script Name   Upgrade_metadata.sas                                                 */ 
/*DBMS          SAS                                                                  */ 
/*Created       01/07/2013 / Updated  - 1/5/2015                                     */ 
/*************************************************************************************/
/*  Define each of these parameters to match your environment                        */ 
/*  =========================================================                        */
/*                                                                                   */
/*                                                                                   */
/* METALIBURI   = REPOSITORY URI FOR LIBRARY NAME TO DDS                             */
/* REPUSER      = USER ID FOR ACCESS TO REPOSITORY                                   */
/* REPPASS      = PASSWORD FOR REPUSER USER ID                                       */
/* REPMACHNAME  = REPOSITORY HOST MACHINE                                            */
/* DDSREPOSNAME = REPOSITORY NAME WHERE DDS IS STORED                                */
/*    note - FOUNDATION is the default                                               */
/* PORT = TCP PORT USED BY SAS Metadata Server i.e.8561                              */
/* METAFOLDERMAIN = METADATA FOLDER PATH FOR DDS_MAIN                                */
/* METAFOLDERREF  = METADATA FOLDER PATH FOR DDS_REFERENCE                           */
/* METAFOLDERALL  = METADATA FOLDER PATH Containing DDS_MAIN + DDS_REFERENCE FOLDERS */
/* DDS_NOTES_ID   = metadata URI for DDS_NOTES folder                                */
/* datastore      = datastore of the DDS;  set to "SAS" where the DDS is stored in   */
/*                SAS datasets otherwise set to "dbms" or no value for DDS stored    */
/*                in other supported datastores such as Oracle.                      */
/* installloc - Windows or Unix Install Location                                     */
/*                                                                                   */
/* **NOTE**                                                                          */
/*  =========================================================                        */                                                                     */
/* Review folder locations and modify to site structure                              */
/*                                                                                   */
/* update_rule option of delete is included in the last step to clean up metadata and*/
/* synchronize with physical DDS tables - MAKE SURE METADATA DELETE IS ACCEPTABLE.   */
/* update_rule=( delete  ) Must include the DELETE rule to remove metadata objects   */  
/*                                                                                   */     
/*************************************************************************************/


options VALIDVARNAME=ANY;

/* --Set Input Variables--                                     */ 
 
%Let METALIBURI   = <LIBRARY_URI>;
%let REPUSER      =%quote(<repository user>);
%let REPPASS      =%quote(<password for repository user>);
%let REPMACHNAME  =%quote(<repository machine>);  
%let DDSREPOSNAME =%quote(Foundation); 
%let PORT         =%quote(<port>);

/* note this is the metadata folder structure as distributed in 4.7, change according */
/* to site setup                                                                         */

%LET METAFOLDERMAIN  =%quote(<metadata_folder-path>/SAS Detail Data Store for Banking/DDS 4.7/DDS_MAIN);
%LET METAFOLDERREF   =%quote(<metadata_folder-path>/SAS Detail Data Store for Banking/DDS 4.7/DDS_REFERENCE);
%LET METAFOLDERALL   =%quote(<metadata_folder-path>/SAS Detail Data Store for Banking/DDS 4.7);

%Let DDS_NOTES_ID=<metadata URI>;  /* metadata uri id of the DDS_NOTES folder              */
%let datastore=SAS;   /* set to SAS if the DDS datastore is SAS otherwise will assume      */
                      /* another database; if another database, set to dbms or no value */
%let installloc=<Windows or Unix Install Location>/bankingdds_4.7/upgrade/metadata; 

/* ---  END OF VARIABLES TO POPULATE  ---                                            */

/* In_Meta - csv file in distributed upgrade package with table and column notes     */
/* used to upgrade the table and column notes                                        */

%let In_Meta=&installloc./Table_Column_Notes.csv;


/* set up meta datastore macro variable based on DDS datastore                       */
%global meta_Tmplt datastore;
%macro ddsdataset();
data _null_;
%if %upcase(&datastore)=SAS %then %do;
call symput("datastore", 'SAS');
call symput("meta_Tmplt", "<Tables/></SASLibrary>");
%end;
%else %do;
call symput("datastore", '');
call symput("meta_Tmplt", "<UsingPackages/></SASLibrary><DatabaseSchema id="""" Name=""""><Tables/></DatabaseSchema>");
%end;
run;
%mend ddsdataset;
%ddsdataset(); 

/*Map files that will be used for parsing metadata xml  
  used by XML engine to get id's (SASLibrary/schema, table, column) produced by proc metadata */
%LET Map_TabNote=&installloc./%trim(&datastore.)TableCommentMap.map;
%LET Map_ColNote=&installloc./%trim(&datastore.)TableColumnCommentMap.map;

 /* DDS_MAIN new table registration and update existing tables with new or changed columns */
 /* input file with list of changed main type tables in hot fix or maintenance             */
%let enttbls="&installloc./entire_tbl_list.sas";

 proc metalib;
            omr (LIBURI   = "&METALIBURI"    /* REPOSITORY ID FOR LIBNAME TO DDS         */
                 USER     = "&REPUSER"      /* USER ID FOR ACCESS TO REPOSITORY          */
                 PASSWORD = "&REPPASS"      /* PASSWORD FOR USER ID                      */
                 IPADDR   = "&REPMACHNAME"  /* REPOSITORY HOST MACHINE                   */
                 REPNAME  = "&DDSREPOSNAME" /* REPOSITORY NAME TO STORE METADATA FOR DDS */
                 PROTOCOL = BRIDGE          /* REQUIRED PROTOCOL                         */
                 PORT     = "&PORT" 
                );

   %include &enttbls;

   folder="&METAFOLDERMAIN";      

    report=(type=detail);  		

*   update_rule=( delete  );  

run;

 /* DDS_REFERENCE new table registration and update existing tables with new or changed columns */
/* input file with list of reference type tables in 4.7                                         */
%let reftbls="&installloc./ref_tbl_list.sas";

  proc metalib;
            omr (
                 LIBURI   = "&METALIBURI"    /* REPOSITORY ID FOR LIBNAME TO DDS         */
                 USER     = "&REPUSER"      /* USER ID FOR ACCESS TO REPOSITORY          */
                 PASSWORD = "&REPPASS"      /* PASSWORD FOR USER ID                      */
                 IPADDR   = "&REPMACHNAME"  /* REPOSITORY HOST MACHINE                   */
                 REPNAME  = "&DDSREPOSNAME" /* REPOSITORY NAME TO STORE METADATA FOR DDS */
                 PROTOCOL = BRIDGE          /* REQUIRED PROTOCOL                         */
                 PORT     = "&PORT"
                );

   %include &reftbls;

    folder="&METAFOLDERREF";  
  
    report=(type=detail);  		
 
*   update_rule=( delete  );
  
run;

/* ----------------------------------------------------------------------------------- */
/* Upgrade Notes - Upload definitions for tables and column to notes metadata object   */
/* ----------------------------------------------------------------------------------- */
/*                                                                                     */
options metaserver=&REPMACHNAME;  
options metaport=&PORT; 
options metaprotocol=bridge;
options metauser="&REPUSER";
options metapass=&REPPASS;

/* Get repository ids and names for tables and columns, notes and the note text      */

filename meta_fl TEMP;  

proc metadata
	REPOS="&DDSREPOSNAME"
	IN="<GetMetadataObjects>
        <Reposid>$METAREPOSITORY</Reposid>
		<Type>SASLibrary</Type>
        <Objects/><Ns>SAS</Ns><Flags>388</Flags>
        <Options>
		<XMLSELECT search=""SASLibrary[@id = '&METALIBURI.']""/>
		<Templates>
			<SASLibrary id="""" Name="""">&meta_Tmplt.
			<PhysicalTable id="""" Name=""""><Documents/><Columns/></PhysicalTable>
			<Column id="""" Name=""""><Documents/></Column>
			<Document id="""" Name="""" ><Notes/></Document>
			<TextStore id="""" Name=""""></TextStore>
			<Document id="""" Name=""""><Notes/></Document>
			<TextStore id="""" Name=""""/>
    	</Templates>
		</Options>
  		</GetMetadataObjects>"
	out=meta_fl; 
run;

/*  Load the ERwin CSV file which contains the comments to load into Notes metadata  */
data WORK.Table_Column_Notes;
   %let _EFIERR_ = 0; /* set the ERROR detection macro variable */
   infile "&In_Meta" delimiter = ',' MISSOVER DSD lrecl=32767 firstobs=2 ;
   informat TABLENAME $40. ;
   informat TABLENOTE $1000. ;
   informat COLNAME $40. ;
   informat COLNOTE $1000. ;
   format TABLENAME $40. ;
   format TABLENOTE $1000. ;
   format COLNAME $40. ;
   format COLNOTE $1000. ;
   input
         TABLENAME $
         TABLENOTE $
         COLNAME $
         COLNOTE $
    ;
   if _ERROR_ then call symputx('_EFIERR_',1);  /* set ERROR detection macro variable */
run;
data Table_Column_Notes;
	set Table_Column_Notes;
	* Remove single/double quote (") from the comment -- replace it with `;
	* Replace "&" with "-" and '"' with '`';
	* Replace "<" with "{" and ">" with "}";;
	TableNote = translate(TableNote, '`-{}`','"&<>''');
	ColNote = translate(ColNote, '`-{}`','"&<>''');
run;

/* Map file to parse the xml output With Table comments */
filename  TabNote "&Map_TabNote.";

/* Create an xml libname using the map */
libname   meta_fl xml xmlmap=TabNote access=READONLY;

/* match existing metadata tables with the input note upgrade data set to use to upload the table definitions */
%let prefx=%trim(&datastore.);
proc sql;
	Create Table work.table_updnotes as
	Select distinct T.TableId,T.TableName,T.TabCommentId, PT.TableNote
	   	From meta_fl.&prefx.Tablenote T, Work.Table_Column_Notes PT
	Where upcase(trim(T.TableName))=upcase(trim(PT.TableName))
	Order by T.TableName;
quit;


/* Map file to parse the xml output With Column comments */
filename  ColNote "&Map_ColNote.";

/* Create a xml libname using the map */
libname   meta_fl xml xmlmap=ColNote access=READONLY;

/* match existing metadata tables and columns with the input note upgrade data to use to upload the column definitions */
proc sql;
	Create Table work.column_note_meta as
	Select T.TableId,T.TableName, PT.TabCommentId, PT.TableNote, T.Columnid, T.ColumnName,
	       T.ColCommentId, T.ColCommentName,
		PT.COLNOTE as ColumnNote
	From meta_fl.&prefx.TableColumnNote T, (select x.TableName, x.TabCommentID, y.TableNote, y.COLNAME, y.COLNOTE 
                                             FROM work.Table_updnotes x, Work.Table_Column_Notes y
                                             where upcase(trim(x.TableName))=upcase(trim(y.TABLENAME))) PT
	Where upcase(trim(T.TableName))=upcase(trim(PT.TableName))
	And T.ColumnName=PT.ColName
	Order by T.TableName,T.ColumnName;
quit;
proc sql;
create table work.col_meta_notes_existing as 
  Select T.TableId,T.TableName, T.TabCommentId, T.TableNote, T.Columnid, T.ColumnName,
	       T.ColCommentId, T.ColCommentName,T.ColumnNote 
		   from work.column_note_meta T
		   where ColCommentId is not missing
		   order by ColCommentId, TableName; 
 create table work.col_meta_notes_missing as	   
  Select T.TableId,T.TableName, T.TabCommentId, T.TableNote, T.Columnid, T.ColumnName,
	       T.ColCommentId, T.ColCommentName,T.ColumnNote 
		   from work.column_note_meta T
		   where ColCommentId is missing
		   order by ColCommentId, TableName;
quit;
data work.table_column_updnotes;
set work.col_meta_notes_existing work.col_meta_notes_missing;
run;

filename tnoteupd TEMP; 

/*Build the xml file that will be used for updating table notes information*/
data _null_;
	length string $10000; *for storing intermediate info;
	file tnoteupd  LRECL=5000; *output file;

	set work.table_updnotes end=EOF;	
		by TableName notsorted;
	/*for first observation create the header line of xml*/
	if _n_ = 1 then	do;
		string= '<UpdateMetadata><Reposid>$METAREPOSITORY</Reposid><Metadata>';
		put string;
	end;
	/*check if note exist in metadata - if it does, update the note */	
	if TabCommentId > "" then	do;
		string= '<TextStore Id="'||trim(left(TabCommentId))||'"'||' StoredText="'||trim(left(TableNote))||'"/>';
		put string;
	end;
	/*if note doesn't exist - new table note will be built during column note update (next step)*/	

	/*Close out tags if it is the last line*/
	if EOF then	do;
		string='</Metadata><NS>SAS</NS><Options/><Flags>268435456</Flags></UpdateMetadata>';
		PUT string;
	end;
run;

filename cnoteupd TEMP; 

/*Build the xml file that will be used for updating column notes information*/
data _null_;
	length string $10000; *for storing intermediate info;
	file cnoteupd  LRECL=5000; *output file;

	set work.table_column_updnotes end=EOF;	
		by TableName notsorted;
	/*for first observation create the header line of xml*/
	if _n_ = 1 then	do;
		string= '<UpdateMetadata><Reposid>$METAREPOSITORY</Reposid><Metadata>';
		put string;
	end;
	/* if the column note exists, update it    */
	if ColCommentId > "" then do;
		string= '<TextStore Id="'||trim(left(ColCommentId))||'"'||' StoredText="'||trim(left(ColumnNote))||'"/>';
		put string;
	end;
	/*if the column note does not exist, and first occurance of table, build the table level tags*/	
	else do;
		if first.TableName then	do;
			string= '<PhysicalTable Id="'||TableId||'"><Columns>';
			put string;
	  	end;

	/*Build Column information Note Naming: <Table Name>.<Column Name>_Note
	  Notes are attached to Documents which are attached to the Notes folder
	*/
		string = '<Column Id="'||ColumnId||'">';
		put string;
	/* need to determine how to attach documents to existing column ??  ID?? try it without id first - maybe update will create id? */
		string = ' <Documents><Document Name="' || trim(left(tablename))||'.' ||trim(ColumnName)
			||'_Note" PublicType="Note" UsageVersion="1000000" TextRole="Note">';  
        put string;    
	/*' Associate the document with notes folder   */ 
		string = '  <Trees><Tree ObjRef="' || "&DDS_NOTES_ID."||'"/></Trees>';
		put string;
		string = ' <Notes> <TextStore Name="'||trim(left(TableName))||'.'
				||trim(ColumnName)||'_Note"  StoredText="' || trim(left(ColumnNote)) 
				||'" TextRole="Note" TextType="PlainText" UsageVersion="1000000">' 
				|| '</TextStore></Notes></Document></Documents></Column>';
		put string;
	/*if last column of the table then build table note and close table tags*/
		if last.TableName then	do;
			put '</Columns>';
	/*Build table note information Note Naming: <Table Name>_Note*/
	/*Notes are attached to documents */
			string = '  <Documents><Document Name="' || trim(left(tablename))||'_Note" PublicType="Note" UsageVersion="1000000" TextRole="Note">';
			put string;

	/* Associate the document with notes folder    */ 
			string = '  <Trees><Tree ObjRef="' || "&DDS_NOTES_ID."||'"/></Trees>';
			put string;
			string = '<Notes> <TextStore Name="' || trim(left(TableName)) 
						     || '_Note"  StoredText="'|| trim(left(TableNote)) 
			         	     || '" UsageVersion="1000000" TextRole="Note" TextType="PlainText">' 
						     ||'</TextStore></Notes></Document></Documents>';
			put string;
			put '</PhysicalTable>';
		end;
	end;
	/*Close out tags if it is the last line*/
	if EOF then	do;
		string='</Metadata><NS>SAS</NS><Options/><Flags>268435456</Flags></UpdateMetadata>';
		PUT string;
	end;
run;

/* upload/update table notes   */
proc metadata
	REPOS="&DDSREPOSNAME."
	in= tnoteupd;
run;
/* upload/update column notes   */
proc metadata
	REPOS="&DDSREPOSNAME."
	in= cnoteupd;
run;
